# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tracking.py"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
import pandas as pd

import pytest

from deepcell_tracking import tracking
from deepcell_tracking import utils
from deepcell_tracking.test_utils import get_annotated_movie


class DummyModel(object):  # pylint: disable=useless-object-inheritance

    def predict(self, data):
        # Grab a random value from the data dict and select batch dim
        if data:
            batches = 1
            frames = 1
            pred_shape = next(iter(data.values())).shape
            cells = pred_shape[1]
        else:
            batches = 0
            frames = 0
            cells = 0

        return np.random.random((batches, cells, cells, frames, 3))


class DummyEncoder(object):  # pylint: disable=useless-object-inheritance

    def predict(self, data):
        # Data should be of the shape (frames, cells, 32, 32, 1)
        # Where frames = number of images in the movie or X.shape[0]
        # and   cells = number of cells (unique IDs) in the movie
        # so grab a random value from the data dict and store correct dim
        if data:
            pred_shape = next(iter(data.values())).shape
            frames = pred_shape[0]  # track length
            cells = pred_shape[1]
        else:
            frames = 0
            cells = 0

        return [np.random.random((frames, cells, 64)),
                np.random.random((frames, cells, 2))]


class TestTracking(object):  # pylint: disable=useless-object-inheritance

    def test_simple(self):
        data_format = 'channels_last'
        frames = 3
        labels_per_frame = 5
        y = get_annotated_movie(img_size=256,
                                labels_per_frame=labels_per_frame,
                                frames=frames,
                                mov_type='sequential', seed=0,
                                data_format=data_format)
        x = np.random.random(y.shape)
        num_objects = len(np.unique(y)) - 1
        model = DummyModel()
        encoder = DummyEncoder()

        _ = tracking.CellTracker(x, y,
                                 tracking_model=model,
                                 neighborhood_encoder=encoder)

        # test data with bad rank
        with pytest.raises(ValueError):
            tracking.CellTracker(
                np.random.random((32, 32, 1)),
                np.random.randint(num_objects, size=(32, 32, 1)),
                tracking_model=model,
                neighborhood_encoder=encoder)

        # test mismatched x and y shape
        with pytest.raises(ValueError):
            tracking.CellTracker(
                np.random.random((3, 32, 32, 1)),
                np.random.randint(num_objects, size=(2, 32, 32, 1)),
                tracking_model=model,
                neighborhood_encoder=encoder)

        # test bad data_format
        with pytest.raises(ValueError):
            tracking.CellTracker(x, y,
                                 tracking_model=model,
                                 neighborhood_encoder=encoder,
                                 data_format='invalid')

    def test_track_cells(self, tmpdir):
        frames = 10
        track_length = 3
        labels_per_frame = 3

        # TODO: test detected divisions
        # TODO: test creating new track

        # TODO: Fix for channels_first
        for data_format in ('channels_last',):  # 'channels_first'):

            y1 = get_annotated_movie(img_size=256,
                                     labels_per_frame=labels_per_frame,
                                     frames=frames,
                                     mov_type='sequential', seed=1,
                                     data_format=data_format)
            y2 = get_annotated_movie(img_size=256,
                                     labels_per_frame=labels_per_frame * 2,
                                     frames=frames,
                                     mov_type='sequential', seed=2,
                                     data_format=data_format)
            y3 = get_annotated_movie(img_size=256,
                                     labels_per_frame=labels_per_frame,
                                     frames=frames,
                                     mov_type='sequential', seed=3,
                                     data_format=data_format)

            y = np.concatenate((y1, y2, y3))

            x = np.random.random(y.shape)

            tracker = tracking.CellTracker(
                x, y,
                tracking_model=DummyModel(),
                neighborhood_encoder=DummyEncoder(),
                track_length=track_length,
                data_format=data_format)

            tracker.track_cells()

            # test tracker.dataframe
            df = tracker.dataframe(cell_type='test-value')
            assert isinstance(df, pd.DataFrame)
            assert 'cell_type' in df.columns  # pylint: disable=E1135

            # test incorrect values in tracker.dataframe
            with pytest.raises(ValueError):
                tracker.dataframe(bad_value=-1)

            # test tracker.postprocess
            tempdir = str(tmpdir)
            path = os.path.join(tempdir, 'postprocess.xyz')
            tracker.postprocess(filename=path)
            post_saved_path = os.path.join(tempdir, 'postprocess.trk')
            assert os.path.isfile(post_saved_path)

            # test tracker.dump
            path = os.path.join(tempdir, 'test.xyz')
            tracker.dump(path)
            dump_saved_path = os.path.join(tempdir, 'test.trk')
            assert os.path.isfile(dump_saved_path)

            # utility tests for loading trk files
            # TODO: move utility tests into utils_test.py

            # test trk_folder_to_trks
            utils.trk_folder_to_trks(tempdir, os.path.join(tempdir, 'all.trks'))
            assert os.path.isfile(os.path.join(tempdir, 'all.trks'))

            # test load_trks
            data = utils.load_trks(post_saved_path)
            assert isinstance(data['lineages'], list)
            assert all(isinstance(d, dict) for d in data['lineages'])
            np.testing.assert_equal(data['X'], tracker.X)
            np.testing.assert_equal(data['y'], tracker.y_tracked)
            # load trks instead of trk
            data = utils.load_trks(os.path.join(tempdir, 'all.trks'))

            # test trks_stats
            utils.trks_stats(os.path.join(tempdir, 'all.trks'))



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for the ISBI data format."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import Counter
from skimage.measure import regionprops

import networkx as nx
import numpy as np
import pandas as pd
import warnings

from deepcell_toolbox import compute_overlap
from deepcell_tracking.utils import load_trks


def trk_to_isbi(track, path=None):
    """Convert a lineage track into an ISBI formatted text file.

    Args:
        track (dict): Cell lineage object.
        path (str): Path to save the .txt file (deprecated).

    Returns:
        pd.DataFrame: DataFrame of ISBI data for each label.
    """
    isbi = []
    for label in track:
        first_frame = min(track[label]['frames'])
        last_frame = max(track[label]['frames'])
        parent = track[label]['parent']
        parent = 0 if parent is None else parent
        if parent:
            parent_frames = track[parent]['frames']
            if parent_frames[-1] != first_frame - 1:
                parent = 0

        isbi_dict = {'Cell_ID': label,
                     'Start': first_frame,
                     'End': last_frame,
                     'Parent_ID': parent}
        isbi.append(isbi_dict)

    if path is not None:
        with open(path, 'w') as text_file:
            for cell in isbi_dict:
                line = '{cell_id} {start} {end} {parent}\n'.format(
                    cell_id=cell['Cell_ID'],
                    start=cell['Start'],
                    end=cell['End'],
                    parent=cell['Parent_ID']
                )
                text_file.write(line)
    df = pd.DataFrame(isbi)
    return df


def contig_tracks(label, batch_info, batch_tracked):
    """Check for contiguous tracks (tracks should only consist of consecutive frames).

    Split one track into two if neccesary

    Args:
        label (int): label of the cell.
        batch_info (dict): a track's lineage info
        batch_tracked (dict): the new image data associated with the lineage.

    Returns:
        tuple(dict, dict): updated batch_info and batch_tracked.
    """
    frame_div_missing = False

    frames = batch_info[label]['frames']

    for i, frame in enumerate(frames):
        # If the next frame is available and contiguous we should move on to
        # the next frame. Otherwise, if the next frame is available and
        # NONcontiguous we should separate this track into two.
        if i + 1 <= len(frames) - 1 and frame + 1 != frames[i + 1]:
            frame_div = batch_info[label].get('frame_div')
            if frame_div is None:
                frame_div_missing = True  # TODO: is this necessary?

            # Create a new track to hold the information from this
            # frame forward and add it to the batch.
            new_label = max(batch_info) + 1
            batch_info[new_label] = {
                'old_label': label,
                'label': new_label,
                'frames': frames[i + 1:],
                'daughters': batch_info[label]['daughters'],
                'frame_div': frame_div,
                'parent': None
            }

            for d in batch_info[new_label]['daughters']:
                batch_info[d]['parent'] = new_label

            for f in frames[i + 1:]:
                batch_tracked[f][batch_tracked[f] == label] = new_label

            # Adjust the info of the current track to vacate the new track info
            batch_info[label]['frames'] = frames[0:i + 1]
            batch_info[label]['daughters'] = []
            batch_info[label]['frame_div'] = None

            break  # Because we are splitting tracks recursively, we stop here

        # If the current frame is the last frame then were done
        # Either the last frame is contiguous and we don't alter batch_info
        # or it's not and it's been made into a new track by the previous
        # iteration of the loop

        if frame_div_missing:
            print('Warning: frame_div is missing')

    return batch_info, batch_tracked


def match_nodes(gt, res):
    """Relabel predicted track to match GT track labels.

    Args:
        gt (np arr): label movie (y) from ground truth .trk file.
        res (np arr): label movie (y) from predicted results .trk file
        threshold (int): threshold value for IoU to count as same cell

    Returns:
        gtcells (np arr): Array of overlapping ids in the gt movie.
        rescells (np arr): Array of overlapping ids in the res movie.

    Raises:
        ValueError: If .
    """
    num_frames = gt.shape[0]
    iou = np.zeros((num_frames, np.max(gt) + 1, np.max(res) + 1))

    # TODO: Compute IOUs only when neccesary
    # If bboxs for true and pred do not overlap with each other, the assignment is immediate
    # Otherwise use pixel-wise IOU to determine which cell is which

    # Regionprops expects one frame at a time
    for frame in range(num_frames):
        gt_frame = gt[frame]
        res_frame = res[frame]

        gt_props = regionprops(np.squeeze(gt_frame.astype('int')))
        gt_boxes = [np.array(gt_prop.bbox) for gt_prop in gt_props]
        gt_boxes = np.array(gt_boxes).astype('double')
        gt_box_labels = [int(gt_prop.label) for gt_prop in gt_props]

        res_props = regionprops(np.squeeze(res_frame.astype('int')))
        res_boxes = [np.array(res_prop.bbox) for res_prop in res_props]
        res_boxes = np.array(res_boxes).astype('double')
        res_box_labels = [int(res_prop.label) for res_prop in res_props]

        overlaps = compute_overlap(gt_boxes, res_boxes)    # has the form [gt_bbox, res_bbox]

        # Find the bboxes that have overlap at all (ind_ corresponds to box number - starting at 0)
        ind_gt, ind_res = np.nonzero(overlaps)

        # frame_ious = np.zeros(overlaps.shape)
        for index in range(ind_gt.shape[0]):

            iou_gt_idx = gt_box_labels[ind_gt[index]]
            iou_res_idx = res_box_labels[ind_res[index]]
            intersection = np.logical_and(gt_frame == iou_gt_idx, res_frame == iou_res_idx)
            union = np.logical_or(gt_frame == iou_gt_idx, res_frame == iou_res_idx)
            iou[frame, iou_gt_idx, iou_res_idx] = intersection.sum() / union.sum()

    gtcells, rescells = np.where(np.nansum(iou, axis=0) >= 1)

    return gtcells, rescells


def txt_to_graph(path, node_key=None):
    """Read the ISBI text file and create a Graph.

    Args:
        path (str): Path to the ISBI text file.
        node_key (dict): Map between gt nodes and result nodes

    Returns:
        networkx.Graph: Graph representation of the ISBI data.

    Raises:
        ValueError: If the Parent_ID is not in any previous frames.
    """
    names = ['Cell_ID', 'Start', 'End', 'Parent_ID']
    df = pd.read_csv(path, header=None, sep=' ', names=names)
    G = isbi_to_graph(df, node_key)
    return G


def isbi_to_graph(df, node_key=None):
    """Create a Graph from DataFrame of ISBI info.

    Args:
        data (pd.DataFrame): DataFrame of ISBI-style info.
        node_key (dict): Map between gt nodes and result nodes

    Returns:
        networkx.Graph: Graph representation of the ISBI data.

    Raises:
        ValueError: If the Parent_ID is not in any previous frames.
    """
    if node_key is not None:
        df[['Cell_ID', 'Parent_ID']] = df[['Cell_ID', 'Parent_ID']].replace(node_key)

    edges = pd.DataFrame()

    all_ids = set()
    single_nodes = set()

    # Add each continuous cell lineage as a set of edges to df
    for _, row in df.iterrows():
        tpoints = np.arange(row['Start'], row['End'] + 1)

        cellids = ['{}_{}'.format(row['Cell_ID'], t) for t in tpoints]

        if len(cellids) == 1:
            single_nodes.add(cellids[0])

        all_ids.update(cellids)

        edges = edges.append(pd.DataFrame({
            'source': cellids[0:-1],
            'target': cellids[1:],
        }))

    attributes = {}

    # Add parent-daughter connections
    for _, row in df[df['Parent_ID'] != 0].iterrows():
        # Assume the parent is in the previous frame.
        parent_frame = row['Start'] - 1
        source = '{}_{}'.format(row['Parent_ID'], parent_frame)

        if source not in all_ids:  # parents should be in the previous frame.
            # parent_frame = df[df['Cell_ID'] == row['Parent_id']]['End']
            # source = '{}_{}'.format(row['Parent_ID'], parent_frame)
            print('skipped parent %s to daughter %s' % (source, row['Cell_ID']))
            continue

        target = '{}_{}'.format(row['Cell_ID'], row['Start'])

        edges = edges.append(pd.DataFrame({
            'source': [source],
            'target': [target]
        }))

        attributes[source] = {'division': True}

    # Create graph
    G = nx.from_pandas_edgelist(edges, source='source', target='target',
                                create_using=nx.DiGraph)
    nx.set_node_attributes(G, attributes)

    # Add all isolates to graph
    for cell_id in single_nodes:
        G.add_node(cell_id)
    return G


def classify_divisions(G_gt, G_res):
    """Compare two graphs and calculate the cell division confusion matrix.

    WARNING: This function will only work if the labels underlying both
    graphs are the same. E.G. the parents only match if the same label
    splits in the same frame - but each movie isn't guaranteed to be labeled
    in the same way (with the same order). Should be used with match_nodes

    Args:
        G_gt (networkx.Graph): Ground truth cell lineage graph.
        G_res (networkx.Graph): Predicted cell lineage graph.

    Returns:
        dict: Diciontary of all division statistics.
    """
    # Identify nodes with parent attribute
    div_gt = [node for node, d in G_gt.nodes(data=True)
              if d.get('division', False)]
    div_res = [node for node, d in G_res.nodes(data=True)
               if d.get('division', False)]

    correct = 0         # Correct division
    incorrect = 0       # Wrong division
    false_positive = 0  # False positive division
    missed = 0          # Missed division

    for node in div_gt:

        pred_gt = list(G_gt.pred[node])
        succ_gt = list(G_gt.succ[node])

        # Check if res node was also called a division
        if node in div_res:
            pred_res = list(G_gt.pred[node])
            succ_res = list(G_res.succ[node])

            # Parents and daughters are the same, perfect!
            if (Counter(pred_gt) == Counter(pred_res) and
                    Counter(succ_gt) == Counter(succ_res)):
                correct += 1

            else:  # what went wrong?
                incorrect += 1
                errors = ['out degree = {}'.format(G_res.out_degree(node))]
                if Counter(succ_gt) != Counter(succ_res):
                    errors.append('daughters mismatch')
                if Counter(pred_gt) != Counter(pred_res):
                    errors.append('parents mismatch')
                if G_res.out_degree(node) == G_gt.out_degree(node):
                    errors.append('gt and res degree equal')
                print(node, '{}.'.format(', '.join(errors)))

            div_res.remove(node)

        else:  # valid division not in results, it was missed
            print('missed node {} division completely'.format(node))
            missed += 1

    # Count any remaining res nodes as false positives
    false_positive += len(div_res)

    return {
        'Correct division': correct,
        'Mismatch division': incorrect,
        'False positive division': false_positive,
        'False negative division': missed,
        'Total divisions': len(div_gt)
    }


def benchmark_division_performance(trk_gt, trk_res, path_gt=None, path_res=None):
    """Compare two related .trk files (one being the GT of the other) and measure
    performance on the the divisions in the GT file.

    Args:
        trk_gt (path): Path to the ground truth .trk file.
        trk_res (path): Path to the predicted results .trk file.
        path_gt (path): Desired destination path for the GT ISBI-style .txt
            file (deprecated).
        path_res (path): Desired destination path for the result ISBI-style
            .txt file (deprecated).

    Returns:
        dict: Dictionary of all division statistics.
    """
    # Identify nodes with parent attribute
    # Load both .trk
    trks = load_trks(trk_gt)
    lineage_gt, _, y_gt = trks['lineages'][0], trks['X'], trks['y']
    trks = load_trks(trk_res)
    lineage_res, _, y_res = trks['lineages'][0], trks['X'], trks['y']

    # Produce ISBI style array to work with
    if path_gt is not None or path_res is not None:
        warnings.warn('The `path_gt` and `path_res` arguments are deprecated.',
                      DeprecationWarning)
    gt = trk_to_isbi(lineage_gt, path_gt)
    res = trk_to_isbi(lineage_res, path_res)

    # Match up labels in GT to Results to allow for direct comparisons
    cells_gt, cells_res = match_nodes(y_gt, y_res)

    if len(np.unique(cells_res)) < len(np.unique(cells_gt)):
        node_key = {r: g for g, r in zip(cells_gt, cells_res)}
        # node_key maps gt nodes onto resnodes so must be applied to gt
        G_res = isbi_to_graph(res, node_key=node_key)
        G_gt = isbi_to_graph(gt)
    else:
        node_key = {g: r for g, r in zip(cells_gt, cells_res)}
        G_res = isbi_to_graph(res)
        G_gt = isbi_to_graph(gt, node_key=node_key)

    div_results = classify_divisions(G_gt, G_res)

    return div_results


def calculate_summary_stats(true_positive,
                            false_positive,
                            false_negative,
                            total_divisions,
                            n_digits=2):
    """Calculate additional summary statistics for tracking performance
    based on results of classify_divisions

    Catch ZeroDivisionError and set to 0 instead

    Args:
        true_positive (int): True positive or "correct divisions"
        false_positive (int): False positives
        false_negative (int): False negatives
        total_divisions (int): Total number of ground truth divisions
        n_digits (int, optional): Number of digits to round to. Default 2.
    """

    _round = lambda x: round(x, n_digits)

    try:
        recall = true_positive / (true_positive + false_negative)
    except ZeroDivisionError:
        recall = 0

    try:
        precision = true_positive / (true_positive + false_positive)
    except ZeroDivisionError:
        precision = 0

    try:
        f1 = 2 * (recall * precision) / (recall + precision)
    except ZeroDivisionError:
        f1 = 0

    try:
        mbc = true_positive / (true_positive + false_negative + false_positive)
    except ZeroDivisionError:
        mbc = 0

    try:
        fraction_miss = (false_negative + false_positive) / total_divisions
    except ZeroDivisionError:
        fraction_miss = 0

    return {
        'Recall': _round(recall),
        'Precision': _round(precision),
        'F1': _round(f1),
        'Mitotic branching correctness': _round(mbc),
        'Fraction missed divisions': _round(fraction_miss)
    }



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tracking_utils"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import os
import io

import numpy as np
import skimage as sk

import pytest

from deepcell_tracking import utils
from deepcell_tracking.test_utils import get_image
from deepcell_tracking.test_utils import get_annotated_image
from deepcell_tracking.test_utils import get_annotated_movie


def get_dummy_data(num_labels=3, batches=2):
    num_labels = 3
    movies = []
    for _ in range(batches):
        movies.append(get_annotated_movie(labels_per_frame=num_labels))

    y = np.stack(movies, axis=0)
    X = np.random.random(y.shape)

    # create dummy lineage
    lineages = {}
    for b in range(X.shape[0]):
        lineages[b] = {}
        for frame in range(X.shape[1]):
            unique_labels = np.unique(y[b, frame])
            unique_labels = unique_labels[unique_labels != 0]
            for unique_label in unique_labels:
                if unique_label not in lineages[b]:
                    lineages[b][unique_label] = {
                        'frames': [frame],
                        'parent': None,
                        'daughters': [],
                        'label': unique_label,
                    }
                else:
                    lineages[b][unique_label]['frames'].append(frame)

    # tracks expect batched data
    data = {'X': X, 'y': y, 'lineages': lineages}
    return data


class TestTrackingUtils(object):

    def test_clean_up_annotations(self):
        img = sk.measure.label(sk.data.binary_blobs(length=256, n_dim=2)) * 3
        img = np.expand_dims(img, axis=-1)
        img = np.expand_dims(img, axis=0)  # time axis
        uid = 1

        cleaned = utils.clean_up_annotations(
            img, uid=uid, data_format='channels_last')
        unique = np.unique(cleaned)
        assert len(np.unique(img)) == len(unique)
        expected = np.arange(len(unique)) + uid - 1
        expected[0] = 0  # background shouldn't get added
        np.testing.assert_equal(expected, unique)

        img = sk.measure.label(sk.data.binary_blobs(length=256, n_dim=2)) * 3
        img = np.expand_dims(img, axis=0)
        img = np.expand_dims(img, axis=1)  # time axis

        cleaned = utils.clean_up_annotations(
            img, uid=uid, data_format='channels_first')
        unique = np.unique(cleaned)
        assert len(np.unique(img)) == len(unique)
        expected = np.arange(len(unique)) + uid - 1
        expected[0] = 0  # background shouldn't get added
        np.testing.assert_equal(expected, unique)

        # Correctness check
        for mov_type in ('random', 'repeated'):
            labels_per_frame = 3
            frames = 3
            movie = get_annotated_movie(img_size=256,
                                        labels_per_frame=labels_per_frame,
                                        frames=frames,
                                        mov_type=mov_type, seed=1,
                                        data_format='channels_last')
            cleaned = utils.clean_up_annotations(movie, uid=uid)
            for frame in range(frames):
                unique = np.unique(cleaned[frame, :, :, 0])
                start = (frame * labels_per_frame) + 1
                end = labels_per_frame * (frame + 1)
                expected = np.arange(start, end + 1, 1, dtype='int32')
                if start != 0:
                    expected = np.append(0, expected)
                np.testing.assert_array_equal(unique, expected)

    def test_count_pairs(self):
        batches = 1
        frames = 2
        classes = 4
        prob = 0.5
        expected = batches * frames * classes * (classes + 1) / prob

        # channels_last
        y = np.random.randint(low=0, high=classes + 1,
                              size=(batches, frames, 30, 30, 1))
        pairs = utils.count_pairs(y, same_probability=prob)
        assert pairs == expected

        # channels_first
        y = np.random.randint(low=0, high=classes + 1,
                              size=(batches, 1, frames, 30, 30))
        pairs = utils.count_pairs(
            y, same_probability=prob, data_format='channels_first')
        assert pairs == expected

    def test_save_trks(self, tmpdir):
        X = get_image(30, 30)
        y = np.random.randint(low=0, high=10, size=X.shape)
        lineage = [dict()]

        tempdir = str(tmpdir)
        with pytest.raises(ValueError):
            badfilename = os.path.join(tempdir, 'x.trk')
            utils.save_trks(badfilename, lineage, X, y)

        filename = os.path.join(tempdir, 'x.trks')
        utils.save_trks(filename, lineage, X, y)
        assert os.path.isfile(filename)

        # test saved tracks can be loaded
        loaded = utils.load_trks(filename)
        assert loaded['lineages'] == lineage
        np.testing.assert_array_equal(X, loaded['X'])
        np.testing.assert_array_equal(y, loaded['y'])

        # test save trks to bytes
        b = io.BytesIO()
        utils.save_trks(b, lineage, X, y)

        # load trks from bytes
        b.seek(0)
        loaded = utils.load_trks(b)
        assert loaded['lineages'] == lineage
        np.testing.assert_array_equal(X, loaded['X'])
        np.testing.assert_array_equal(y, loaded['y'])

    def test_save_trk(self, tmpdir):
        X = get_image(30, 30)
        y = np.random.randint(low=0, high=10, size=X.shape)
        lineage = [dict()]

        tempdir = str(tmpdir)
        with pytest.raises(ValueError):
            badfilename = os.path.join(tempdir, 'x.trks')
            utils.save_trk(badfilename, lineage, X, y)

        with pytest.raises(ValueError):
            utils.save_trk('x.trk', [{}, {}], X, y)

        filename = os.path.join(tempdir, 'x.trk')
        utils.save_trk(filename, lineage, X, y)
        assert os.path.isfile(filename)

        # test saved tracks can be loaded
        loaded = utils.load_trks(filename)
        assert loaded['lineages'] == lineage
        np.testing.assert_array_equal(X, loaded['X'])
        np.testing.assert_array_equal(y, loaded['y'])

        # test save trks to bytes
        b = io.BytesIO()
        utils.save_trk(b, lineage, X, y)

        # load trks from bytes
        b.seek(0)
        loaded = utils.load_trks(b)
        assert loaded['lineages'] == lineage
        np.testing.assert_array_equal(X, loaded['X'])
        np.testing.assert_array_equal(y, loaded['y'])

    def test_load_trks(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'bad-lineage.trk')
        X = get_image(30, 30)
        y = np.random.randint(low=0, high=10, size=X.shape)
        lineage = [dict()]

        utils.save_track_data(filename=filename,
                              lineages=lineage,
                              raw=X,
                              tracked=y,
                              lineage_name='bad-lineage.json')

        with pytest.raises(ValueError):
            utils.load_trks(filename)

    def test_normalize_adj_matrix(self):
        frames = 3
        # 2 cells in each frame
        adj = np.zeros((frames, 2, 2))
        for i in range(2):
            adj[:, i, i] = 1

        normalized = utils.normalize_adj_matrix(adj)

        # TODO: check accuracy of normalized tensor.

        # also normalize batches
        batched_adj = np.stack([adj, adj], axis=0)
        batched_normalized = utils.normalize_adj_matrix(batched_adj)

        # each batch is the same, should be normalized consistently
        for b in range(batched_normalized.shape[0]):
            np.testing.assert_array_equal(
                batched_normalized[b],
                normalized)

        # Should fail with too large inputs
        with pytest.raises(ValueError):
            utils.normalize_adj_matrix(np.zeros((32,) * 2))

        # Should fail with too small inputs
        with pytest.raises(ValueError):
            utils.normalize_adj_matrix(np.zeros((32,) * 5))

    def test_get_max_cells(self):
        labels_per_frame = 5
        frames = 2
        expected_max = labels_per_frame * 2
        y1 = get_annotated_movie(img_size=256,
                                 labels_per_frame=labels_per_frame,
                                 frames=frames,
                                 mov_type='sequential', seed=1,
                                 data_format='channels_last')
        y2 = get_annotated_movie(img_size=256,
                                 labels_per_frame=labels_per_frame * 2,
                                 frames=frames,
                                 mov_type='sequential', seed=2,
                                 data_format='channels_last')
        y3 = get_annotated_movie(img_size=256,
                                 labels_per_frame=labels_per_frame,
                                 frames=frames,
                                 mov_type='sequential', seed=3,
                                 data_format='channels_last')
        y = np.concatenate((y1, y2, y3))
        calculated_max = utils.get_max_cells(y)
        assert expected_max == calculated_max

    def test_relabel_sequential_lineage(self):
        # create dummy movie
        image1 = get_annotated_image(num_labels=1, sequential=False)
        image2 = get_annotated_image(num_labels=2, sequential=False)
        movie = np.stack([image1, image2], axis=0)

        # create dummy lineage
        lineage = {}
        for frame in range(movie.shape[0]):
            unique_labels = np.unique(movie[frame])
            unique_labels = unique_labels[unique_labels != 0]
            for unique_label in unique_labels:
                lineage[unique_label] = {
                    'frames': [frame],
                    'parent': None,
                    'daughters': [],
                    'label': unique_label,
                }
        # assign parents and daughters
        parent_label = np.unique(movie[0])
        parent_label = parent_label[parent_label != 0][0]

        daughter_labels = np.unique(movie[1])
        daughter_labels = [d for d in daughter_labels if d]

        lineage[parent_label]['daughters'] = daughter_labels
        for d in daughter_labels:
            lineage[d]['parent'] = parent_label

        # relabel the movie and lineage
        new_movie, new_lineage = utils.relabel_sequential_lineage(movie, lineage)
        new_parent_label = int(np.unique(new_movie[np.where(movie == parent_label)]))

        # test parent is relabeled
        assert new_parent_label == 1  # sequential should start at 1
        assert new_lineage[new_parent_label]['frames'] == lineage[parent_label]['frames']
        assert new_lineage[new_parent_label]['parent'] is None
        assert new_lineage[new_parent_label]['label'] == new_parent_label

        # test daughters are relabeled
        new_daughter_labels = new_lineage[new_parent_label]['daughters']
        assert len(new_daughter_labels) == 2

        for d in new_daughter_labels:
            old_label = int(np.unique(movie[np.where(new_movie == d)]))
            assert new_lineage[d]['frames'] == lineage[old_label]['frames']
            assert new_lineage[d]['parent'] == new_parent_label
            assert new_lineage[d]['label'] == d
            assert not new_lineage[d]['daughters']

    def test_is_valid_lineage(self):
        image1 = get_annotated_image(num_labels=1, sequential=False)
        image2 = get_annotated_image(num_labels=2, sequential=False)
        movie = np.stack([image1, image2], axis=0)

        # create dummy lineage
        lineage = {}
        for frame in range(movie.shape[0]):
            unique_labels = np.unique(movie[frame])
            unique_labels = unique_labels[unique_labels != 0]
            for unique_label in unique_labels:
                lineage[unique_label] = {
                    'frames': [frame],
                    'parent': None,
                    'daughters': [],
                    'label': unique_label,
                }

        # assign parents and daughters
        parent_label = np.unique(movie[0])
        parent_label = parent_label[parent_label != 0][0]

        daughter_labels = np.unique(movie[1])
        daughter_labels = [d for d in daughter_labels if d]

        lineage[parent_label]['daughters'] = daughter_labels
        for d in daughter_labels:
            lineage[d]['parent'] = parent_label

        assert utils.is_valid_lineage(movie, lineage)

        # a cell's frames should match the label array
        bad_lineage = copy.deepcopy(lineage)
        bad_lineage[parent_label]['frames'].append(1)
        assert not utils.is_valid_lineage(movie, bad_lineage)

        # cell in lineage but not in movie is invalid
        max_label = np.max(movie)
        bad_label = np.max(movie) + 1
        bad_lineage = copy.deepcopy(lineage)
        bad_lineage[bad_label] = bad_lineage[max_label]
        assert not utils.is_valid_lineage(movie, bad_lineage)

        # cell in movie but not in lineage is invalid
        bad_movie = copy.deepcopy(movie)
        bad_movie[0, 0, 0] = bad_label
        assert not utils.is_valid_lineage(bad_movie, lineage)

        # a daughter's frames should start immediately
        # after the parent's last frame
        # (not strictly true in data, but required for GNN models)
        bad_frame = 0
        bad_movie = copy.deepcopy(movie)
        bad_movie[bad_frame, 0, 0] = daughter_labels[0]
        bad_lineage = copy.deepcopy(lineage)
        bad_lineage[daughter_labels[0]]['frames'] = [bad_frame]
        assert not utils.is_valid_lineage(bad_movie, bad_lineage)

        # daughter not in lineage is invalid
        bad_lineage = copy.deepcopy(lineage)
        bad_movie = copy.deepcopy(movie)
        daughter_idx = np.where(movie == lineage[parent_label]['daughters'][0])
        bad_movie[daughter_idx] = bad_label
        bad_lineage[parent_label]['daughters'][0] = bad_label
        assert not utils.is_valid_lineage(bad_movie, bad_lineage)

        # daughter not in movie is invalid
        bad_lineage = copy.deepcopy(lineage)
        bad_lineage[parent_label]['daughters'][0] = bad_label
        bad_lineage[bad_label] = bad_lineage[bad_lineage[parent_label]['daughters'][1]]
        assert not utils.is_valid_lineage(movie, bad_lineage)

        # test daughter frames are empty
        bad_lineage = copy.deepcopy(lineage)
        bad_lineage[daughter_labels[0]]['frames'] = []
        assert not utils.is_valid_lineage(movie, bad_lineage)

        # parent ID > daughters ID is OK
        new_parent = max(np.unique(movie)) + 2
        relabeled_movie = np.where(movie == parent_label, new_parent, movie)
        relabeled_lineage = copy.deepcopy(lineage)
        relabeled_lineage[new_parent] = lineage[parent_label]
        del relabeled_lineage[parent_label]
        for daughter in relabeled_lineage[new_parent]['daughters']:
            if relabeled_lineage[daughter]['parent'] == parent_label:
                relabeled_lineage[daughter]['parent'] = new_parent
        assert utils.is_valid_lineage(relabeled_movie, relabeled_lineage)

    def test_get_image_features(self):
        num_labels = 3
        y = get_annotated_image(num_labels=num_labels, sequential=True)
        y = np.expand_dims(y, axis=-1)
        X = np.random.random(y.shape)

        appearance_dim = 16
        distance_threshold = 64
        features = utils.get_image_features(X, y, appearance_dim)

        # test appearance
        appearances = features['appearances']
        expected_shape = (num_labels, appearance_dim, appearance_dim, X.shape[-1])
        assert appearances.shape == expected_shape

        # test centroids
        centroids = features['centroids']
        expected_shape = (num_labels, 2)
        assert centroids.shape == expected_shape

        # test centroids
        centroids = features['centroids']
        expected_shape = (num_labels, 2)
        assert centroids.shape == expected_shape

        # test morphologies
        morphologies = features['morphologies']
        expected_shape = (num_labels, 3)
        assert morphologies.shape == expected_shape

        # test labels
        labels = features['labels']
        expected_shape = (num_labels,)
        assert labels.shape == expected_shape
        np.testing.assert_array_equal(labels, np.array(list(range(1, num_labels + 1))))

    def test_trks_stats(self):
        # Test bad extension
        with pytest.raises(ValueError):
            utils.trks_stats('bad-extension.npz')

        # No inputs
        with pytest.raises(ValueError):
            utils.trks_stats()

        data = get_dummy_data()
        stats = utils.trks_stats(**data)
        assert isinstance(stats, dict)



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for utility functions for the ISBI data format."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import os
import tarfile
import tempfile
import json

import networkx as nx
import numpy as np
import pandas as pd

from deepcell_tracking import isbi_utils
from deepcell_tracking.test_utils import get_annotated_movie


class TestIsbiUtils(object):

    def test_trk_to_isbi(self, tmpdir):
        # start with dummy lineage
        # convert to ISBI array
        # validate array

        track = {}
        # first cell, skips frame 3 but divides in frame 4
        track[1] = {
            'frames': [0, 1, 2, 4],  # skipped a frame
            'daughters': [2, 3],
            'frame_div': 4,
            'parent': None,
            'label': 1,
        }
        track[2] = {
            'frames': [5],
            'daughters': [],
            'frame_div': None,
            'parent': 1,
            'label': 2,
        }
        track[3] = {
            'frames': [5],
            'daughters': [4],  # parent not in previous frame
            'frame_div': 5,
            'parent': 1,
            'label': 3,
        }
        track[4] = {
            'frames': [7],
            'daughters': [],
            'frame_div': None,
            'parent': 3,
            'label': 4,
        }
        df = isbi_utils.trk_to_isbi(track)

        expected = [{'Cell_ID': 1, 'Start': 0, 'End': 4, 'Parent_ID': 0},
                    {'Cell_ID': 2, 'Start': 5, 'End': 5, 'Parent_ID': 1},
                    {'Cell_ID': 3, 'Start': 5, 'End': 5, 'Parent_ID': 1},
                    {'Cell_ID': 4, 'Start': 7, 'End': 7, 'Parent_ID': 0}]
        expected_df = pd.DataFrame(expected)
        assert df.equals(expected_df)

    def test_txt_to_graph(self, tmpdir):
        # cell_id, start, end, parent_id
        rows = [
            (1, 0, 3, 0),  # cell 1 is in all 3 frames
            (2, 0, 2, 0),  # cell 2 is not in the last frame
            (3, 3, 3, 2),  # cell 3 is a daughter of 2
            (4, 3, 3, 2),  # cell 4 is a daughter of 2
            (5, 3, 3, 4),  # cell 5 is a daughter of 4, ignored bad frame value
        ]
        text_file = os.path.join(str(tmpdir), 'test_txt_to_graph.txt')
        with open(text_file, 'wb') as f:
            # write the file
            for row in rows:
                line = '{} {} {} {}{}'.format(
                    row[0], row[1], row[2], row[3], os.linesep)
                f.write(line.encode())

            f.flush()  # save the file

        # read the file
        G = isbi_utils.txt_to_graph(text_file)
        for row in rows:
            node_ids = ['{}_{}'.format(row[0], t)
                        for t in range(row[1], row[2] + 1)]

            for node_id in node_ids:
                assert node_id in G

            if row[3]:  # should have a division
                daughter_id = '{}_{}'.format(row[0], row[1])
                parent_id = '{}_{}'.format(row[3], row[1] - 1)
                if G.has_node(parent_id):
                    assert G.nodes[parent_id]['division'] is True
                    assert G.has_edge(parent_id, daughter_id)
                else:
                    assert not G.in_degree(daughter_id)

    def test_isbi_to_graph(self):
        # cell_id, start, end, parent_id
        data = [{'Cell_ID': 1, 'Start': 0, 'End': 3, 'Parent_ID': 0},
                {'Cell_ID': 2, 'Start': 0, 'End': 2, 'Parent_ID': 0},
                {'Cell_ID': 3, 'Start': 3, 'End': 3, 'Parent_ID': 2},
                {'Cell_ID': 4, 'Start': 3, 'End': 3, 'Parent_ID': 2},
                {'Cell_ID': 5, 'Start': 3, 'End': 3, 'Parent_ID': 4}]
        df = pd.DataFrame(data)
        G = isbi_utils.isbi_to_graph(df)
        for d in data:
            node_ids = ['{}_{}'.format(d["Cell_ID"], t)
                        for t in range(d["Start"], d["End"] + 1)]

            for node_id in node_ids:
                assert node_id in G

            if d["Parent_ID"]:  # should have a division
                daughter_id = '{}_{}'.format(d["Cell_ID"], d["Start"])
                parent_id = '{}_{}'.format(d["Parent_ID"], d["Start"] - 1)
                if G.has_node(parent_id):
                    assert G.nodes[parent_id]['division'] is True
                    assert G.has_edge(parent_id, daughter_id)
                else:
                    assert not G.in_degree(daughter_id)

    def test_classify_divisions(self):
        G = nx.DiGraph()
        G.add_edge('1_0', '1_1')
        G.add_edge('1_1', '1_2')
        G.add_edge('1_2', '1_3')

        G.add_edge('2_0', '2_1')
        G.add_edge('2_1', '2_2')

        # node 2 divides into 3 and 4 in frame 3
        G.add_edge('2_2', '3_3')
        G.add_edge('2_2', '4_3')
        G.nodes['2_2']['division'] = True

        G.add_edge('4_3', '4_4')  # another division in frame 4
        G.nodes['4_3']['division'] = True

        H = G.copy()

        H.nodes['1_3']['division'] = True  # False Positive
        H.nodes['4_3']['division'] = False  # False Negative

        # force an incorrect division
        G.add_edge('3_3', '5_4')  # another division in frame 4
        G.nodes['3_3']['division'] = True
        H.nodes['3_3']['division'] = True

        stats = isbi_utils.classify_divisions(G, H)

        assert stats['Correct division'] == 1  # the only correct one
        assert stats['False positive division'] == 1  # node 1_3
        assert stats['False negative division'] == 1  # node 4_3
        assert stats['Mismatch division'] == 1  # node 3_3
        assert stats['Total divisions'] == 3

    def test_contig_tracks(self):
        # test already contiguous
        frames = 5
        track = {
            1: {
                'label': 1,
                'frames': [0, 1, 2],
                'daughters': [2, 3],
                'parent': None,
            },
            2: {
                'label': 2,
                'frames': [3, 4],
                'daughters': [],
                'parent': 1
            },
            3: {
                'label': 3,
                'frames': [3, 4],
                'daughters': [],
                'parent': 1
            }
        }
        original_track = copy.copy(track)
        original_daughters = original_track[1]['daughters']
        y = np.random.randint(0, 4, size=(frames, 40, 40, 1))
        new_track, _ = isbi_utils.contig_tracks(1, track, y)
        assert original_track == new_track

        # test non-contiguous
        track = copy.copy(original_track)
        track[1]['frames'].append(4)
        new_track, _ = isbi_utils.contig_tracks(1, track, y)

        assert len(new_track) == len(original_track) + 1
        assert new_track[1]['frames'] == original_track[1]['frames']
        daughters = new_track[max(new_track)]['daughters']
        assert daughters == original_daughters
        for d in daughters:
            assert new_track[d]['parent'] == max(new_track)

    def test_match_nodes(self):
        # creat dummy movie to test against
        labels_per_frame = 5
        frames = 3
        y1 = get_annotated_movie(img_size=256,
                                 labels_per_frame=labels_per_frame,
                                 frames=frames,
                                 mov_type='repeated', seed=1,
                                 data_format='channels_last')
        # test same movie
        gtcells, rescells = isbi_utils.match_nodes(y1, y1)
        for gt_cell, res_cell in zip(gtcells, rescells):
            assert gt_cell == res_cell

        # test different movie (with known values)
        y2 = get_annotated_movie(img_size=256,
                                 labels_per_frame=labels_per_frame,
                                 frames=frames,
                                 mov_type='sequential', seed=1,
                                 data_format='channels_last')
        gtcells, rescells = isbi_utils.match_nodes(y1, y2)

        assert len(rescells) == len(gtcells)
        for loc, gt_cell in enumerate(np.unique(gtcells)):
            # because movies have the same first frame, every
            # iteration of unique values should match original label
            assert gt_cell == rescells[loc * 3]

    def test_benchmark_division_performance(self, tmpdir):
        trk_gt = os.path.join(str(tmpdir), 'test_benchmark_gt.trk')
        trk_res = os.path.join(str(tmpdir), 'test_benchmark_res.trk')

        # Generate lineage data
        tracks_gt = {1: {'label': 1, 'frames': [1, 2], 'daughters': [],
                     'capped': False, 'frame_div': None, 'parent': 3},
                     2: {'label': 2, 'frames': [1, 2], 'daughters': [],
                     'capped': False, 'frame_div': None, 'parent': 3},
                     3: {'label': 3, 'frames': [0], 'daughters': [1, 2],
                     'capped': False, 'frame_div': 1, 'parent': None}}
        X_gt = []
        # Generate tracked movie
        y_gt = get_annotated_movie(img_size=256,
                                   labels_per_frame=3,
                                   frames=3,
                                   mov_type='sequential', seed=0,
                                   data_format='channels_last')
        # Let results be same as ground truth
        tracks_res = tracks_gt
        X_res = []
        y_res = y_gt

        # Save gt and res data to .trk files
        with tarfile.open(trk_gt, 'w:gz') as trks:
            # disable auto deletion and close/delete manually
            # to resolve double-opening issue on Windows.
            with tempfile.NamedTemporaryFile('w', delete=False) as lineage:
                json.dump(tracks_gt, lineage, indent=4)
                lineage.flush()
                lineage.close()
                trks.add(lineage.name, 'lineage.json')
                os.remove(lineage.name)

            with tempfile.NamedTemporaryFile(delete=False) as raw:
                np.save(raw, X_gt)
                raw.flush()
                raw.close()
                trks.add(raw.name, 'raw.npy')
                os.remove(raw.name)

            with tempfile.NamedTemporaryFile(delete=False) as tracked:
                np.save(tracked, y_gt)
                tracked.flush()
                tracked.close()
                trks.add(tracked.name, 'tracked.npy')
                os.remove(tracked.name)

        with tarfile.open(trk_res, 'w:gz') as trks:
            # disable auto deletion and close/delete manually
            # to resolve double-opening issue on Windows.
            with tempfile.NamedTemporaryFile('w', delete=False) as lineage:
                json.dump(tracks_res, lineage, indent=4)
                lineage.flush()
                lineage.close()
                trks.add(lineage.name, 'lineage.json')
                os.remove(lineage.name)

            with tempfile.NamedTemporaryFile(delete=False) as raw:
                np.save(raw, X_res)
                raw.flush()
                raw.close()
                trks.add(raw.name, 'raw.npy')
                os.remove(raw.name)

            with tempfile.NamedTemporaryFile(delete=False) as tracked:
                np.save(tracked, y_res)
                tracked.flush()
                tracked.close()
                trks.add(tracked.name, 'tracked.npy')
                os.remove(tracked.name)

        expected = {'Correct division': 1, 'Mismatch division': 0,
                    'False positive division': 0, 'False negative division': 0,
                    'Total divisions': 1}
        results = isbi_utils.benchmark_division_performance(trk_gt, trk_res)
        assert results == expected



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for testing deepcell-tracking"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import skimage as sk


def get_image(img_h=300, img_w=300):
    bias = np.random.rand(img_w, img_h) * 64
    variance = np.random.rand(img_w, img_h) * (255 - 64)
    img = np.random.rand(img_w, img_h) * variance + bias
    return img


def get_annotated_image(img_size=256, num_labels=3, sequential=True, seed=1):
    np.random.seed(seed)
    num_labels_act = False
    trial = 0
    while num_labels != num_labels_act:
        if trial > 10:
            raise Exception('Labels have merged despite 10 different random seeds.'
                            ' Increase image size or reduce the number of labels')
        im = np.zeros((img_size, img_size))
        points = img_size * np.random.random((2, num_labels))
        im[(points[0]).astype(np.int), (points[1]).astype(np.int)] = 1
        im = sk.filters.gaussian(im, sigma=5)
        blobs = im > 0.7 * im.mean()
        all_labels, num_labels_act = sk.measure.label(blobs, return_num=True)
        if num_labels != num_labels_act:
            seed += 1
            np.random.seed(seed)
            trial += 1

    if not sequential:
        labels_in_frame = np.unique(all_labels)
        for label in range(num_labels):
            curr_label = label + 1
            new_label = np.random.randint(1, num_labels * 100)
            while new_label in labels_in_frame:
                new_label = np.random.randint(1, num_labels * 100)
            labels_in_frame = np.append(labels_in_frame, new_label)
            label_loc = np.where(all_labels == curr_label)
            all_labels[:, :][label_loc] = new_label

    return all_labels.astype('int32')


def get_annotated_movie(img_size=256, labels_per_frame=3, frames=3,
                        mov_type='sequential', seed=1,
                        data_format='channels_last'):
    if mov_type in ('sequential', 'repeated'):
        sequential = True
    elif mov_type == 'random':
        sequential = False
    else:
        raise ValueError('mov_type must be one of "sequential", '
                         '"repeated" or "random"')

    if data_format == 'channels_last':
        channel_axis = -1
    else:
        channel_axis = 0

    y = []
    while len(y) < frames:
        _y = get_annotated_image(img_size=img_size, num_labels=labels_per_frame,
                                 sequential=sequential, seed=seed)
        y.append(_y)
        seed += 1

    y = np.stack(y, axis=0)  # expand to 3D

    if mov_type == 'sequential':
        for frame in range(frames):
            if frame == 0:
                new_label = labels_per_frame
                continue
            for label in range(labels_per_frame):
                curr_label = label + 1
                new_label += 1
                label_loc = np.where(y[frame, :, :] == curr_label)
                y[frame, :, :][label_loc] = new_label

    y = np.expand_dims(y, axis=channel_axis)

    return y.astype('int32')



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A cell tracking class capable of extending labels across sequential frames."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import json
import logging
import os
import pathlib
import tarfile
import tempfile
import timeit

import numpy as np
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from skimage.measure import regionprops

import networkx as nx
import pandas as pd

from deepcell_tracking.utils import clean_up_annotations
from deepcell_tracking.utils import get_max_cells
from deepcell_tracking.utils import normalize_adj_matrix
from deepcell_tracking.utils import get_image_features
from deepcell_tracking.utils import save_trk


class CellTracker(object):  # pylint: disable=useless-object-inheritance
    """Solves the linear assingment problem to build a cell lineage graph.

    Args:
        movie (np.array): raw time series movie of cells.
        annotation (np.array): the labeled cell movie.
        model (keras.Model): tracking model to determine if two cells are the
            same, different, or parent/daughter.
        features (list): list of strings for the features to use.
        crop_dim (int): crop size for the appearance feature.
        death (float): paramter used to fill the death matrix in the LAP,
            (top right of the cost matrix).
        birth (float): paramter used to fill the birth matrix in the LAP,
            (bottom left of the cost matrix).
        division (float): probability threshold for assigning daughter cells.
        distance_threshold (int): maximum distance to compare cells with the model.
        track_length (int): the track length used for the model.
        neighborhood_scale_size (int): neighborhood feature size to pass to the
            model.
        neighborhood_true_size (int): original size of the neighborhood feature
            which will be scaled down to neighborhood_scale_size.
        dtype (str): data type for features, can be 'float32', 'float16', etc.
        data_format (str): determines the order of the channel axis,
            one of 'channels_first' and 'channels_last'.
    """

    def __init__(self,
                 movie,
                 annotation,
                 tracking_model,
                 neighborhood_encoder=None,
                 distance_threshold=64,
                 appearance_dim=32,
                 death=0.99,
                 birth=0.99,
                 division=0.9,
                 track_length=5,
                 embedding_axis=0,
                 dtype='float32',
                 data_format='channels_last'):

        if not len(movie.shape) == 4 or not len(annotation.shape) == 4:
            raise ValueError('Input data and labels must be rank 4 (frames, x, y, channels). '
                             'Got rank {} (X) and rank {} (y).'.format(
                                 len(movie.shape), len(annotation.shape)))

        if not movie.shape[:-1] == annotation.shape[:-1]:
            raise ValueError('Input data and labels should have the same shape'
                             ' except for the channel dimension.  Got {} and '
                             '{}'.format(movie.shape, annotation.shape))

        if data_format not in {'channels_first', 'channels_last'}:
            raise ValueError('The `data_format` argument must be one of '
                             '"channels_first", "channels_last". Received: ' +
                             str(data_format))

        self.X = copy.copy(movie)
        self.y = copy.copy(annotation)
        self.tracks = {}

        self.neighborhood_encoder = neighborhood_encoder
        self.tracking_model = tracking_model
        self.distance_threshold = distance_threshold
        self.appearance_dim = appearance_dim
        self.death = death
        self.birth = birth
        self.division = division
        self.dtype = dtype
        self.track_length = track_length
        self.embedding_axis = embedding_axis

        self.a_matrix = []
        self.c_matrix = []
        self.assignments = []

        self.data_format = data_format
        self.channel_axis = 0 if data_format == 'channels_first' else -1
        self.time_axis = 1 if data_format == 'channels_first' else 0
        self.logger = logging.getLogger(str(self.__class__.__name__))

        # Clean up annotations
        self.y = clean_up_annotations(self.y, data_format=self.data_format)

        # Accounting for 0 (background) label with 0-indexing for tracks
        self.id_to_idx = {}  # int: int mapping
        self.idx_to_id = {}  # (frame, cell_idx): cell_id mapping

        # Establish features for every instance of every cell in the movie
        adj_matrices, appearances, morphologies, centroids = self._est_feats()

        # Compute embeddings for every instance of every cell in the movie
        embeddings = self._get_neighborhood_embeddings(
            appearances=appearances,
            morphologies=morphologies,
            centroids=centroids,
            adj_matrices=adj_matrices)

        # TODO: immutable dict for safety? these values should never change.
        self.features = {
            'embedding': embeddings,
            'centroid': centroids,
        }

    def _get_frame(self, tensor, frame):
        """Helper function for fetching a frame of a tensor.

        Useful for avoiding duplication of the data_format conditional.

        Args:
            tensor (np.array): The 3D tensor to slice.
            frame (int): The frame to slice out of the tensor.

        Returns:
            np.array: the 2D slice of the 3D tensor.
        """
        if self.data_format == 'channels_first':
            return tensor[:, frame]
        return tensor[frame]

    def _get_cells_in_frame(self, frame):
        """Find the labels of cells in the given frame.

        Args:
            frame (int): Frame of interest.

        Returns:
            list: All cell labels in the frame.
        """
        cells = np.unique(self._get_frame(self.y, frame))
        cells = np.delete(cells, np.where(cells == 0))  # remove the background
        return list(cells)

    def _est_feats(self):
        """
        Extract the relevant features from the label movie
        Appearance, morphologies, centroids, and adjacency matrices
        """
        max_cells = get_max_cells(self.y)
        n_frames = self.X.shape[0]
        n_channels = self.X.shape[-1]

        appearances = np.zeros((n_frames,
                                max_cells,
                                self.appearance_dim,
                                self.appearance_dim,
                                n_channels), dtype=np.float32)

        morphologies = np.zeros((n_frames, max_cells, 3), dtype=np.float32)

        centroids = np.zeros((n_frames, max_cells, 2), dtype=np.float32)

        adj_matrix = np.zeros((n_frames, max_cells, max_cells),
                              dtype=np.float32)

        for frame in range(n_frames):

            frame_features = get_image_features(
                self.X[frame], self.y[frame],
                appearance_dim=self.appearance_dim)

            for cell_idx, cell_id in enumerate(frame_features['labels']):
                self.id_to_idx[cell_id] = cell_idx
                self.idx_to_id[(frame, cell_idx)] = cell_id

            num_tracks = len(frame_features['labels'])
            centroids[frame, :num_tracks] = frame_features['centroids']
            morphologies[frame, :num_tracks] = frame_features['morphologies']
            appearances[frame, :num_tracks] = frame_features['appearances']

            cent = centroids[frame]
            distance = cdist(cent, cent, metric='euclidean') < self.distance_threshold
            adj_matrix[frame] = distance.astype('float32')

        # Normalize adj matrix
        norm_adj_matrices = normalize_adj_matrix(adj_matrix)

        return norm_adj_matrices, appearances, morphologies, centroids

    def _get_neighborhood_embeddings(self, appearances, morphologies,
                                     centroids, adj_matrices):
        """Compute the embeddings using the neighborhood encoder"""
        # Build input dictionary for neighborhood encoder model
        inputs = {
            'encoder_app_input': appearances,
            'encoder_morph_input': morphologies,
            'encoder_centroid_input': centroids,
            'encoder_adj_input': adj_matrices,
        }

        # TODO: current model doesnt organize outputs according to ordered list
        #       patching with embedding_axis
        embeddings = self.neighborhood_encoder.predict(inputs)[self.embedding_axis]
        embeddings = np.array(embeddings)
        return embeddings

    def _validate_feature_name(self, feature_name):
        if feature_name not in self.features:
            raise ValueError('{} is an invalid feature name. '
                             'Use one of embedding or centroid'.format(
                                 feature_name))

    def _get_feature(self, frame, cell_id, feature_name='embedding'):
        """Get the feature for a cell in the frame"""
        self._validate_feature_name(feature_name)
        cell_idx = self.id_to_idx[cell_id]
        return self.features[feature_name][frame, cell_idx, :]

    def _get_frame_features(self, frame, feature_name='embedding'):
        """Get the feature for the specified cells in a frame"""
        self._validate_feature_name(feature_name)

        cells_in_frame = self._get_cells_in_frame(frame)

        frame_features = {}
        for cell_id in cells_in_frame:
            f = self._get_feature(frame, cell_id, feature_name=feature_name)
            frame_features[cell_id] = f
        return frame_features

    def _create_new_track(self, frame, old_label):
        """
        This function creates new tracks
        """
        track_id = len(self.tracks)
        new_label = track_id + 1
        embedding = self._get_feature(frame, old_label, feature_name='embedding')
        centroid = self._get_feature(frame, old_label, feature_name='centroid')

        embedding = np.expand_dims(embedding, axis=0)
        centroid = np.expand_dims(centroid, axis=0)

        self.tracks[track_id] = {
            'label': new_label,
            'frames': [frame],
            'frame_labels': [old_label],
            'daughters': [],
            'capped': False,
            'frame_div': None,
            'parent': None,
            'embedding': embedding,
            'centroid': centroid
        }

        if frame > 0 and np.any(self._get_frame(self.y, frame) == new_label):
            raise Exception('new_label already in annotated frame and frame > 0')

        if self.data_format == 'channels_first':
            self.y[:, frame][self.y[:, frame] == old_label] = new_label
        else:
            self.y[frame][self.y[frame] == old_label] = new_label

    def _initialize_tracks(self):
        """Intialize the tracks. Tracks are stored in a dictionary.
        """
        frame = 0  # initial frame
        cell_ids = self._get_cells_in_frame(frame)

        for cell_id in cell_ids:
            self._create_new_track(frame, cell_id)

        # Start a tracked label array
        # TODO: This could be a source of error!
        # we are assigning a pointer not instantiating a new object
        self.y_tracked = self.y[[frame]].astype('int32')

    def _fetch_tracked_features(self, before_frame=None, feature_name='embedding'):
        """Get feature data from each tracked frame less than before_frame.

        Args:
            before_frame (int, optional): The maximum frame to from which to
                fetch feature data.
            feature_name (str): Name of feature to fetch from tracked data.

        Returns:
            dict: dictionary of feature name to np.array of feature data.
        """
        self._validate_feature_name(feature_name)

        if before_frame is None:
            before_frame = self.X.shape[0] + 1  # all frames

        track_valid_frames = ((n, [f for f in d['frames'] if f < before_frame])
                              for n, d in self.tracks.items())
        tracks_with_frames = [(n, f) for n, f in track_valid_frames if len(f) > 0]

        tracked_features = {}
        for i, (n, valid_frames) in enumerate(tracks_with_frames):
            frame_dict = {frame: j for j, frame in enumerate(valid_frames)}
            frames = valid_frames[-self.track_length:]

            if len(frames) != self.track_length:
                # Pad the the frames with the last frame if not enough
                num_missing = self.track_length - len(frames)
                frames = frames + [frames[-1]] * num_missing

            # Get the feature data from the identified frames
            fetched = self.tracks[n][feature_name][[frame_dict[f] for f in frames]]
            tracked_features[i] = fetched

        return tracked_features

    def _build_cost_matrix(self, assignment_matrix):
        """Build the full cost matrix based on the assignment_matrix.

        Args:
            assignment_matrix (np.array): assignment_matrix.

        Returns:
            numpy.array: cost matrix.
        """
        # Initialize cost matrix
        num_tracks, num_cells = assignment_matrix.shape
        cost_matrix = np.zeros((num_tracks + num_cells,) * 2, dtype=self.dtype)

        # assignment matrix - top left
        cost_matrix[0:num_tracks, 0:num_cells] = assignment_matrix

        # birth matrix - bottom left
        birth_diagonal = np.array([self.birth] * num_cells)
        birth_matrix = np.zeros((num_cells, num_cells), dtype=self.dtype)
        birth_matrix = np.diag(birth_diagonal) + np.ones(birth_matrix.shape)
        birth_matrix = birth_matrix - np.eye(num_cells)
        cost_matrix[num_tracks:, 0:num_cells] = birth_matrix

        # death matrix - top right
        death_matrix = np.ones((num_tracks, num_tracks), dtype=self.dtype)
        death_matrix = self.death * np.eye(num_tracks) + death_matrix
        death_matrix = death_matrix - np.eye(num_tracks)
        cost_matrix[0:num_tracks, num_cells:] = death_matrix

        # mordor matrix - bottom right
        cost_matrix[num_tracks:, num_cells:] = assignment_matrix.T
        return cost_matrix

    def _get_cost_matrix(self, frame):
        """Use the model predictions to build an assignment matrix to be solved.

        Args:
            frame (int): The frame with cells to assign.

        Returns:
            tuple: the assignment matrix and the predictions used to build it.
        """
        inputs = {}
        relevant_tracks = []
        for feature_name in self.features:
            # Get the embeddings for previously tracked cells
            current_feature = self._fetch_tracked_features(
                before_frame=frame, feature_name=feature_name)

            # Get the embeddings for the current frame
            future_feature = self._get_frame_features(
                frame=frame, feature_name=feature_name)

            if not relevant_tracks:
                for track_id in current_feature:
                    relevant_tracks.append(track_id)

            # Convert from dict to arrays
            current_feature_arr = np.stack([
                current_feature[k] for k in current_feature
            ], axis=1)  # time axis already included
            future_feature_arr = np.stack([
                future_feature[k] for k in future_feature
            ], axis=0)

            # Add time dimension
            future_feature_arr = np.expand_dims(future_feature_arr, axis=0)

            # Add batch dimension
            current_feature_arr = np.expand_dims(current_feature_arr, axis=0)
            future_feature_arr = np.expand_dims(future_feature_arr, axis=0)

            # Add feature to inputs
            # model expects "current_embeddings" but feature_name is "embedding"
            # TODO: this name is hardcoded based on a model from deepcell-tf
            inputs['current_{}s'.format(feature_name)] = current_feature_arr
            inputs['future_{}s'.format(feature_name)] = future_feature_arr

        t = timeit.default_timer()

        # Perform inference
        predictions = self.tracking_model.predict(inputs)
        predictions = predictions[0, 0, ...]  # Remove the batch and time dimension
        assignment_matrix = 1 - predictions[..., 0]

        for track_id in relevant_tracks:
            if self.tracks[track_id]['capped']:
                assignment_matrix[track_id, :] = 1

        self.a_matrix.append(predictions)

        # Assemble full cost matrix
        cost_matrix = self._build_cost_matrix(assignment_matrix)
        self.logger.debug('Built cost matrix for frame %s in %s s.',
                          frame, timeit.default_timer() - t)
        self.c_matrix.append(cost_matrix)

        predictions_dict = {}
        predictions_dict['predictions'] = predictions
        predictions_dict['track_ids'] = relevant_tracks

        return cost_matrix, predictions_dict

    def _update_tracks(self, assignments, frame, predictions):
        """Update the graph based on the assignment matrix for the frame.

        Use the assignment matrix to determine if each cell in the frame.
        belongs to a track or is a daughter of a track, and updates the graph
        accordingly.

        Args:
            assignments (np.array): completed assignment matrix used to assign
                cells to existing tracks.
            frame (int): the frame of cells to assign.
            predictions (dict): dictionary of trackID-cellID combination,
                and the probability they are the same cell.
        """
        t = timeit.default_timer()
        cells_in_frame = self._get_cells_in_frame(frame)

        # Number of lables present in the current frame (needed to build cost matrix)
        y_tracked_update = np.zeros((1, self.y.shape[1], self.y.shape[2], 1), dtype='int32')

        self.assignments.append(assignments)

        for a in range(assignments.shape[0]):
            track, cell = assignments[a]

            # map the index from the LAP assignment to the cell label
            try:
                cell_id = cells_in_frame[cell]
            except IndexError:
                # cell has "died" or is a shadow assignment
                # no assignment should be made
                continue

            cell_embedding = self._get_feature(frame, cell_id, feature_name='embedding')
            cell_centroid = self._get_feature(frame, cell_id, feature_name='centroid')

            cell_embedding = np.expand_dims(cell_embedding, axis=0)
            cell_centroid = np.expand_dims(cell_centroid, axis=0)

            if track in self.tracks:  # Add cell and frame to track
                self.tracks[track]['frames'].append(frame)
                self.tracks[track]['frame_labels'].append(cell_id)
                self.tracks[track]['embedding'] = np.concatenate([
                    self.tracks[track]['embedding'],
                    cell_embedding
                ], axis=0)
                self.tracks[track]['centroid'] = np.concatenate([
                    self.tracks[track]['centroid'],
                    cell_centroid
                ], axis=0)

                # Labels and indices differ by 1
                y_tracked_update[self.y[[frame]] == cell_id] = track + 1
                self.y[frame][self.y[frame] == cell_id] = track + 1

            else:  # Create a new track if there was a birth
                self._create_new_track(frame, cell_id)
                new_track_id = max(self.tracks)
                new_label = new_track_id + 1
                self.logger.info('Created new track for cell %s.', new_label)

                # See if the new track has a parent
                parent = self._get_parent(frame, cell_id, predictions)
                if parent is not None:
                    self.logger.info('Detected division! Cell %s is daughter '
                                     'of cell %s.', new_label, parent + 1)
                    self.tracks[new_track_id]['parent'] = parent
                    self.tracks[parent]['daughters'].append(new_track_id)
                else:
                    self.tracks[new_track_id]['parent'] = None

                y_tracked_update[self.y[[frame]] == new_label] = new_track_id + 1
                self.y[frame][self.y[frame] == new_label] = new_track_id + 1

        # Check and make sure cells that divided did not get assigned to the same cell
        for track in range(len(self.tracks)):
            if not self.tracks[track]['daughters']:
                continue  # Filter out cells that have not divided

            # Cap tracks for any divided cells
            if not self.tracks[track]['capped']:
                self.tracks[track]['frame_div'] = int(frame)
                self.tracks[track]['capped'] = True

            try:
                frame_idx = self.tracks[track]['frames'].index(frame)
            except ValueError:
                continue  # Filter out tracks that are not in the frame

            # Create new track
            new_track_id = len(self.tracks)
            new_label = new_track_id + 1
            self._create_new_track(frame, self.tracks[track]['frame_labels'][-1])
            self.tracks[new_track_id]['parent'] = track

            # Remove features and frame from old track
            del self.tracks[track]['frames'][frame_idx]
            del self.tracks[track]['frame_labels'][frame_idx]
            self.tracks[track]['embedding'] = np.delete(
                self.tracks[track]['embedding'],
                frame_idx, axis=0
            )
            self.tracks[track]['centroid'] = np.delete(
                self.tracks[track]['centroid'],
                frame_idx, axis=0
            )
            self.tracks[track]['daughters'].append(new_track_id)

            # Change y_tracked_update
            old_label = self.tracks[track]['label']
            y_tracked_update[self.y[[frame]] == old_label] = new_label
            # TODO: Having y and y_tracked is redundant. only one should be changed
            self.y[frame][self.y[frame] == old_label] = new_label

        # Update the tracked label array
        self.y_tracked = np.concatenate([self.y_tracked, y_tracked_update], axis=0)
        self.logger.debug('Updated tracks for frame %s in %s s.',
                          frame, timeit.default_timer() - t)

    def _get_parent(self, frame, cell, predictions):
        """Searches the tracks for the parent of a given cell.

        Args:
            frame (int): the frame the cell appears in.
            cell (int): the label of the cell in the frame.
            predictions (dict): dictionary of trackID-cellID combination,
                and the probability they are the same cell.

        Returns:
            int: The parent cell's id or None if no parent exists.
        """
        parent_id = None
        max_prob = self.division

        track_ids = predictions['track_ids']
        predictions = predictions['predictions']

        for track_idx, track_id in enumerate(track_ids):
            # Make sure capped tracks can't be assigned parents
            if self.tracks[track_id]['capped']:
                continue

            for cell_idx in range(predictions.shape[1]):
                cell_id = self.idx_to_id[(frame, cell_idx)]

                if cell_id == cell:
                    # Do not call a newly-appeared sibling of "cell" a parent
                    if self.tracks[track_id]['frames'] == [frame]:
                        continue

                    # probability cell is part of the track
                    prob = predictions[track_idx, cell_idx, 2]

                    if prob > max_prob:
                        parent_id, max_prob = track_id, prob

        return parent_id

    def _track_frame(self, frame):
        """Inner function for tracking each frame"""
        t = timeit.default_timer()
        self.logger.info('Tracking frame %s', frame)

        cost_matrix, predictions = self._get_cost_matrix(frame)

        row_ind, col_ind = linear_sum_assignment(cost_matrix)
        assignments = np.stack([row_ind, col_ind], axis=1)

        self._update_tracks(assignments, frame, predictions)
        self.logger.info('Tracked frame %s in %s s.',
                         frame, timeit.default_timer() - t)

    def track_cells(self):
        """Tracks all of the cells in every frame.
        """
        start = timeit.default_timer()
        self._initialize_tracks()

        for frame in range(1, self.X.shape[self.time_axis]):
            self._track_frame(frame)

        self.logger.info('Tracked all %s frames in %s s.',
                         self.X.shape[self.time_axis],
                         timeit.default_timer() - start)

    def _track_review_dict(self):
        def process(key, track_item):
            if track_item is None:
                return track_item
            if key == 'daughters':
                return list(map(lambda x: x + 1, track_item))
            elif key == 'parent':
                return track_item + 1
            else:
                return track_item

        track_keys = ['label', 'frames', 'daughters', 'capped', 'frame_div', 'parent']

        return {'tracks': {track['label']: {key: process(key, track[key]) for key in track_keys}
                           for _, track in self.tracks.items()},
                'X': self.X,
                'y': self.y,
                'y_tracked': self.y_tracked}

    def dataframe(self, **kwargs):
        """Returns a dataframe of the tracked cells with lineage.
        Uses only the cell labels not the ids.
        _track_cells must be called first!
        """
        # possible kwargs are extra_columns
        extra_columns = ['cell_type', 'set', 'part', 'montage']
        track_columns = ['label', 'daughters', 'frame_div']

        incorrect_args = set(kwargs) - set(extra_columns)
        if incorrect_args:
            raise ValueError('Invalid argument {}'.format(incorrect_args.pop()))

        # filter extra_columns by the ones we passed in
        extra_columns = [c for c in extra_columns if c in kwargs]

        # extra_columns are the same for every row, cache the values
        extra_column_vals = [kwargs[c] for c in extra_columns if c in kwargs]

        # fill the dataframe
        data = []
        for cell_id, track in self.tracks.items():
            data.append(extra_column_vals + [track[c] for c in track_columns])
        dataframe = pd.DataFrame(data, columns=extra_columns + track_columns)

        # daughters contains track_id not labels
        dataframe['daughters'] = dataframe['daughters'].apply(
            lambda d: [self.tracks[x]['label'] for x in d])

        return dataframe

    def postprocess(self, filename=None, time_excl=9):
        """Use graph postprocessing to eliminate false positive division errors
        using a graph-based detection method. False positive errors are when a
        cell is noted as a daughter of itself before the actual division occurs.
        If a filename is passed, save the state of the cell tracker to a .trk
        ('track') file. time_excl is the minimum number of frames expected to
        exist between legitimate divisions
        """

        # Load data
        track_review_dict = self._track_review_dict()

        # Prep data
        tracked = track_review_dict['y_tracked'].astype('uint16')
        lineage = track_review_dict['tracks']

        # Identify false positives (FPs)
        G = self._track_to_graph(lineage)
        FPs = self._flag_false_pos(G, time_excl)
        FPs_candidates = sorted(FPs.items(), key=lambda v: int(v[0].split('_')[1]))
        FPs_sorted = self._review_candidate_nodes(FPs_candidates)

        # If FPs exist, use the results to correct
        while len(FPs_sorted) != 0:

            lineage, tracked = self._remove_false_pos(lineage, tracked, FPs_sorted[0])
            G = self._track_to_graph(lineage)
            FPs = self._flag_false_pos(G, time_excl)
            FPs_candidates = sorted(FPs.items(), key=lambda v: int(v[0].split('_')[1]))
            FPs_sorted = self._review_candidate_nodes(FPs_candidates)

        # Make sure the assignment is correct
        track_review_dict['y_tracked'] = tracked
        track_review_dict['tracks'] = lineage

        # Save information to a track file file if requested
        if filename is not None:
            self.dump(filename, track_review_dict)

        return track_review_dict

    def dump(self, filename, track_review_dict=None):
        """Writes the state of the cell tracker to a .trk ('track') file.
        Includes raw & tracked images, and a lineage.json for parent/daughter
        information.
        """
        if not track_review_dict:
            track_review_dict = self._track_review_dict()

        filename = pathlib.Path(filename)

        if filename.suffix != '.trk':
            filename = filename.with_suffix('.trk')

        filename = str(filename)

        save_trk(filename=filename,
                 lineage=track_review_dict['tracks'],
                 raw=track_review_dict['X'],
                 tracked=track_review_dict['y_tracked'])

    def _track_to_graph(self, tracks):
        """Create a graph from the lineage information"""
        Dattr = {}
        edges = pd.DataFrame()

        for L in tracks.values():
            # Calculate node ids
            cellid = ['{}_{}'.format(L['label'], f) for f in L['frames']]
            # Add edges from cell ids
            edges = edges.append(pd.DataFrame({'source': cellid[0:-1],
                                               'target': cellid[1:]}))

            # Collect any division attributes
            if L['frame_div'] is not None:
                Dattr['{}_{}'.format(L['label'], L['frame_div'] - 1)] = {'division': True}

            # Create any daughter-parent edges
            if L['parent'] is not None:
                source = '{}_{}'.format(L['parent'], min(L['frames']) - 1)
                target = '{}_{}'.format(L['label'], min(L['frames']))
                edges = edges.append(pd.DataFrame({'source': [source],
                                                   'target': [target]}))

        G = nx.from_pandas_edgelist(edges, source='source', target='target')
        nx.set_node_attributes(G, Dattr)
        return G

    def _flag_false_pos(self, G, time_excl):
        """Examine graph for false positive nodes
        """

        # TODO: Current implementation may eliminate some divisions at the edge of the frame -
        #       Further research needed

        # Identify false positive nodes
        node_fix = []
        for g in (G.subgraph(c) for c in nx.connected_components(G)):
            div_nodes = [n for n, d in g.nodes(data=True) if d.get('division')]
            if len(div_nodes) > 1:
                for nd in div_nodes:
                    if g.degree(nd) == 2:
                        # Check how close suspected FP is to other known divisions

                        keep_div = True
                        for div_nd in div_nodes:
                            if div_nd != nd:
                                time_spacing = abs(int(nd.split('_')[1]) -
                                                   int(div_nd.split('_')[1]))
                                # If division is sufficiently far away
                                # we should exclude it from FP list
                                if time_spacing > time_excl:
                                    keep_div = False

                        if keep_div is True:
                            node_fix.append(nd)

        # Add supplementary information for each false positive
        D = {}
        for node in node_fix:
            D[node] = {
                'false positive': node,
                'neighbors': list(G.neighbors(node)),
                'connected lineages': set([int(node.split('_')[0])
                                           for node in nx.node_connected_component(G, node)])
            }

        return D

    def _review_candidate_nodes(self, FPs_candidates):
        """ review candidate false positive nodes and remove any errant degree 2 nodes.
        """
        FPs_presort = {}
        # review candidate false positive nodes and remove any errant degree 2 nodes
        for candidate_node in FPs_candidates:
            node = candidate_node[0]
            node_info = candidate_node[1]

            neighbors = []  # structure will be [(neighbor1, frame), (neighbor2,frame)]
            for neighbor in node_info['neighbors']:
                neighbor_label = int(neighbor.split('_')[0])
                neighbor_frame = int(neighbor.split('_')[1])
                neighbors.append((neighbor_label, neighbor_frame))

            # if this cell only exists in one frame (and then it divides) but its 2 neighbors
            # both exist in the same frame it will be a degree 2 node but not be a false positive
            if neighbors[0][1] != neighbors[1][1]:
                FPs_presort[node] = node_info

        FPs_sorted = sorted(FPs_presort.items(), key=lambda v: int(v[0].split('_')[1]))

        return FPs_sorted

    def _remove_false_pos(self, lineage, tracked, FP_info):
        """ Remove nodes that have been identified as false positive divisions.
        """
        node = FP_info[0]
        node_info = FP_info[1]

        fp_label = int(node.split('_')[0])
        fp_frame = int(node.split('_')[1])

        neighbors = []  # structure will be [(neighbor1, frame), (neighbor2,frame)]
        for neighbor in node_info['neighbors']:
            neighbor_label = int(neighbor.split('_')[0])
            neighbor_frame = int(neighbor.split('_')[1])
            neighbors.append((neighbor_label, neighbor_frame))

        # Verify that the FP node only 2 neighbors - 1 before it and one after it
        if len(neighbors) == 2:
            # order the neighbors such that the time (frame order) is respected
            if neighbors[0][1] > neighbors[1][1]:
                temp = neighbors[0]
                neighbors[0] = neighbors[1]
                neighbors[1] = temp

            # Decide which labels to extend and which to remove

            # Neighbor_1 has same label as fp - the actual division hasnt occurred yet
            if fp_label == neighbors[0][0]:
                # The model mistakenly identified a division before the actual division occurred
                label_to_remove = neighbors[1][0]
                label_to_extend = neighbors[0][0]

                # Give all of the errant divisions information to the correct track
                lineage[label_to_extend]['frames'].extend(lineage[label_to_remove]['frames'])
                lineage[label_to_extend]['daughters'] = lineage[label_to_remove]['daughters']
                lineage[label_to_extend]['frame_div'] = lineage[label_to_remove]['frame_div']

                # Adjust the parent information for the actual daughters
                daughter_labels = lineage[label_to_remove]['daughters']
                for daughter in daughter_labels:
                    lineage[daughter]['parent'] = lineage[label_to_remove]['parent']

                # Remove the errant node from the annotated images
                channel = 0  # These images should only have one channel
                for frame in lineage[label_to_remove]['frames']:
                    label_loc = np.where(tracked[frame, :, :, channel] == label_to_remove)
                    tracked[frame, :, :, channel][label_loc] = label_to_extend

                # Remove the errant node from the lineage
                del lineage[label_to_remove]

            # Neighbor_2 has same label as fp - the actual division ocurred &
            # the model mistakenly allowed another
            # elif fp_label == neighbors[1][0]:
                # The model mistakenly identified a division after
                # the actual division occurred
                # label_to_remove = fp_label

            # Neither neighbor has same label as fp - the actual division
            # ocurred & the model mistakenly allowed another
            else:
                # The model mistakenly identified a division after the actual division occurred
                label_to_remove = fp_label
                label_to_extend = neighbors[1][0]

                # Give all of the errant divisions information to the correct track
                lineage[label_to_extend]['frames'] = \
                    lineage[fp_label]['frames'] + lineage[label_to_extend]['frames']
                lineage[label_to_extend]['parent'] = lineage[fp_label]['parent']

                # Adjust the parent information for the actual daughter
                parent_label = lineage[fp_label]['parent']
                for d_idx, daughter in enumerate(lineage[parent_label]['daughters']):
                    if daughter == fp_label:
                        lineage[parent_label]['daughters'][d_idx] = label_to_extend

                # Remove the errant node from the annotated images
                channel = 0  # These images should only have one channel
                for frame in lineage[label_to_remove]['frames']:
                    label_loc = np.where(tracked[frame, :, :, channel] == label_to_remove)
                    tracked[frame, :, :, channel][label_loc] = label_to_extend

                # Remove the errant node
                del lineage[label_to_remove]

        else:
            self.logger.error('Error: More than 2 neighbor nodes')

        return lineage, tracked



# Copyright 2016-2022 The Van Valen Lab at the California Institute of
# Technology (Caltech), with support from the Paul Allen Family Foundation,
# Google, & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

from codecs import open
from setuptools import setup
from setuptools import find_packages


here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(here, 'README.md'), 'r', 'utf-8') as f:
    readme = f.read()


VERSION = '0.5.7'
NAME = 'DeepCell_Tracking'
DESCRIPTION = 'Tracking cells and lineage with deep learning.'
LICENSE = 'LICENSE'
AUTHOR = 'Van Valen Lab'
AUTHOR_EMAIL = 'vanvalenlab@gmail.com'
URL = 'https://github.com/vanvalenlab/deepcell-tracking'
DOWNLOAD_URL = ('https://github.com/vanvalenlab/'
                'deepcell-tracking/tarball/{}'.format(VERSION))


setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      url=URL,
      download_url=DOWNLOAD_URL,
      license=LICENSE,
      install_requires=['networkx>=2.1',
                        'numpy',
                        'pandas',
                        'scipy',
                        'scikit-image>=0.14.5',
                        'deepcell-toolbox~=0.11.2'
                        ],
      extras_require={
          'tests': ['pytest<6',
                    'pytest-pep8',
                    'pytest-cov',
                    'pytest-mock']},
      long_description=readme,
      long_description_content_type='text/markdown',
      packages=find_packages(),
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9'])



# Copyright 2016-2022 Van Valen Lab at California Institute of Technology
# (Caltech), with support from the Paul Allen Family Foundation, Google,
# & National Institutes of Health (NIH) under Grant U24CA224309-01.
# All rights reserved.
#
# Licensed under a modified Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.github.com/vanvalenlab/deepcell-tracking/LICENSE
#
# The Work provided may be used for non-commercial academic purposes only.
# For any other use of the Work, including commercial use, please contact:
# vanvalenlab@gmail.com
#
# Neither the name of Caltech nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for tracking cells"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import io
import json
import os
import re
import tarfile
import tempfile
import warnings

import numpy as np

from scipy.spatial.distance import cdist

from skimage.measure import regionprops
from skimage.segmentation import relabel_sequential

from deepcell_toolbox.utils import resize


def clean_up_annotations(y, uid=None, data_format='channels_last'):
    """Relabels every frame in the label matrix.

    Args:
        y (np.array): annotations to relabel sequentially.
        uid (int, optional): starting ID to begin labeling cells.
        data_format (str): determines the order of the channel axis,
            one of 'channels_first' and 'channels_last'.

    Returns:
        np.array: Cleaned up annotations.
    """
    y = y.astype('int32')
    time_axis = 1 if data_format == 'channels_first' else 0
    num_frames = y.shape[time_axis]

    all_uniques = []
    for f in range(num_frames):
        cells = np.unique(y[:, f] if data_format == 'channels_first' else y[f])
        cells = np.delete(cells, np.where(cells == 0))
        all_uniques.append(cells)

    # The annotations need to be unique across all frames
    uid = sum(len(x) for x in all_uniques) + 1 if uid is None else uid
    for frame, unique_cells in zip(range(num_frames), all_uniques):
        y_frame = y[:, frame] if data_format == 'channels_first' else y[frame]
        y_frame_new = np.zeros(y_frame.shape)
        for cell_label in unique_cells:
            y_frame_new[y_frame == cell_label] = uid
            uid += 1
        if data_format == 'channels_first':
            y[:, frame] = y_frame_new
        else:
            y[frame] = y_frame_new
    return y


def count_pairs(y, same_probability=0.5, data_format='channels_last'):
    """Compute number of training samples needed to observe all cell pairs.

    Args:
        y (np.array): 5D tensor of cell labels.
        same_probability (float): liklihood that 2 cells are the same.
        data_format (str): determines the order of the channel axis,
            one of 'channels_first' and 'channels_last'.

    Returns:
        int: the total pairs needed to sample to see all possible pairings.
    """
    total_pairs = 0
    zaxis = 2 if data_format == 'channels_first' else 1
    for b in range(y.shape[0]):
        # count the number of cells in each image of the batch
        cells_per_image = []
        for f in range(y.shape[zaxis]):
            if data_format == 'channels_first':
                num_cells = len(np.unique(y[b, :, f, :, :]))
            else:
                num_cells = len(np.unique(y[b, f, :, :, :]))
            cells_per_image.append(num_cells)

        # Since there are many more possible non-self pairings than there
        # are self pairings, we want to estimate the number of possible
        # non-self pairings and then multiply that number by two, since the
        # odds of getting a non-self pairing are 50%, to find out how many
        # pairs we would need to sample to (statistically speaking) observe
        # all possible cell-frame pairs. We're going to assume that the
        # average cell is present in every frame. This will lead to an
        # underestimate of the number of possible non-self pairings, but it
        # is unclear how significant the underestimate is.
        average_cells_per_frame = sum(cells_per_image) // y.shape[zaxis]
        non_self_cellframes = (average_cells_per_frame - 1) * y.shape[zaxis]
        non_self_pairings = non_self_cellframes * max(cells_per_image)

        # Multiply cell pairings by 2 since the
        # odds of getting a non-self pairing are 50%
        cell_pairings = non_self_pairings // same_probability
        # Add this batch cell-pairings to the total count
        total_pairs += cell_pairings
    return total_pairs


def load_trks(filename):
    """Load a trk/trks file.

    Args:
        filename (str or BytesIO): full path to the file including .trk/.trks
            or BytesIO object with trk file data

    Returns:
        dict: A dictionary with raw, tracked, and lineage data.
    """
    if isinstance(filename, io.BytesIO):
        kwargs = {'fileobj': filename}
    else:
        kwargs = {'name': filename}

    with tarfile.open(mode='r', **kwargs) as trks:

        # numpy can't read these from disk...
        with io.BytesIO() as array_file:
            array_file.write(trks.extractfile('raw.npy').read())
            array_file.seek(0)
            raw = np.load(array_file)

        with io.BytesIO() as array_file:
            array_file.write(trks.extractfile('tracked.npy').read())
            array_file.seek(0)
            tracked = np.load(array_file)

        # trks.extractfile opens a file in bytes mode, json can't use bytes.
        try:
            trk_data = trks.getmember('lineages.json')
        except KeyError:
            try:
                trk_data = trks.getmember('lineage.json')
            except KeyError:
                raise ValueError('Invalid .trk file, no lineage data found.')

        lineages = json.loads(trks.extractfile(trk_data).read().decode())
        lineages = lineages if isinstance(lineages, list) else [lineages]

        # JSON only allows strings as keys, so convert them back to ints
        for i, tracks in enumerate(lineages):
            lineages[i] = {int(k): v for k, v in tracks.items()}

    return {'lineages': lineages, 'X': raw, 'y': tracked}


def trk_folder_to_trks(dirname, trks_filename):
    """Compiles a directory of trk files into one trks_file.

    Args:
        dirname (str): full path to the directory containing multiple trk files.
        trks_filename (str): desired filename (the name should end in .trks).
    """
    lineages = []
    raw = []
    tracked = []

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    file_list = os.listdir(dirname)
    file_list_sorted = sorted(file_list, key=alphanum_key)

    for filename in file_list_sorted:
        trk = load_trks(os.path.join(dirname, filename))
        lineages.append(trk['lineages'][0])  # this is loading a single track
        raw.append(trk['X'])
        tracked.append(trk['y'])

    file_path = os.path.join(os.path.dirname(dirname), trks_filename)

    save_trks(file_path, lineages, raw, tracked)


def save_trks(filename, lineages, raw, tracked):
    """Saves raw, tracked, and lineage data from multiple movies into one trks_file.

    Args:
        filename (str or io.BytesIO): full path to the final trk files or bytes object
            to save the data to
        lineages (list): a list of dictionaries saved as a json.
        raw (np.array): raw images data.
        tracked (np.array): annotated image data.

    Raises:
        ValueError: filename does not end in ".trks".
    """
    ext = os.path.splitext(str(filename))[-1]
    if not isinstance(filename, io.BytesIO) and ext != '.trks':
        raise ValueError('filename must end with `.trks`. Found %s' % filename)

    save_track_data(filename=filename,
                    lineages=lineages,
                    raw=raw,
                    tracked=tracked,
                    lineage_name='lineages.json')


def save_trk(filename, lineage, raw, tracked):
    """Saves raw, tracked, and lineage data for one movie into a trk_file.

    Args:
        filename (str or io.BytesIO): full path to the final trk files or bytes
            object to save the data to
        lineages (list or dict): a list of a single dictionary or a single
            lineage dictionary
        raw (np.array): raw images data.
        tracked (np.array): annotated image data.

    Raises:
        ValueError: filename does not end in ".trks".
    """
    ext = os.path.splitext(str(filename))[-1]
    if not isinstance(filename, io.BytesIO) and ext != '.trk':
        raise ValueError('filename must end with `.trk`. Found %s' % filename)

    # Check that lineages is a dictionary or list of length 1
    if isinstance(lineage, list):
        if len(lineage) > 1:
            raise ValueError('For trk file, lineages must be a dictionary '
                             'or list with a single dictionary')
        else:
            lineage = lineage[0]

    save_track_data(filename=filename,
                    lineages=lineage,
                    raw=raw,
                    tracked=tracked,
                    lineage_name='lineage.json')


def save_track_data(filename, lineages, raw, tracked, lineage_name):
    """Base function for saving tracking data as either trk or trks

    Args:
        filename (str or io.BytesIO): full path to the final trk files or bytes object
            to save the data to
        lineages (list or dict): a list of a single dictionary or a single lineage dictionarys
        raw (np.array): raw images data.
        tracked (np.array): annotated image data.
        lineage_name (str): Filename for the lineage file in the tarfile, either 'lineages.json'
            or 'lineage.json'
    """

    if isinstance(filename, io.BytesIO):
        kwargs = {'fileobj': filename}
    else:
        kwargs = {'name': filename}

    with tarfile.open(mode='w:gz', **kwargs) as trks:
        # disable auto deletion and close/delete manually
        # to resolve double-opening issue on Windows.
        with tempfile.NamedTemporaryFile('w', delete=False) as lineages_file:
            json.dump(lineages, lineages_file, indent=4)
            lineages_file.flush()
            lineages_file.close()
            trks.add(lineages_file.name, lineage_name)
            os.remove(lineages_file.name)

        with tempfile.NamedTemporaryFile(delete=False) as raw_file:
            np.save(raw_file, raw)
            raw_file.flush()
            raw_file.close()
            trks.add(raw_file.name, 'raw.npy')
            os.remove(raw_file.name)

        with tempfile.NamedTemporaryFile(delete=False) as tracked_file:
            np.save(tracked_file, tracked)
            tracked_file.flush()
            tracked_file.close()
            trks.add(tracked_file.name, 'tracked.npy')
            os.remove(tracked_file.name)


def trks_stats(filename=None, X=None, y=None, lineages=None):
    """For a given trks_file, find the Number of cell tracks,
       the Number of frames per track, and the Number of divisions.

    Args:
        filename (str): full path to a trks file.

    Raises:
        ValueError: filename is not a .trk or .trks file.
    """
    if filename:
        ext = os.path.splitext(filename)[-1].lower()
        if ext not in {'.trks', '.trk'}:
            raise ValueError(
                '`trks_stats` expects a .trk or .trks but found a {}'.format(ext))

        training_data = load_trks(filename)
        X = training_data['X']
        y = training_data['y']
        lineages = training_data['lineages']

    if not filename and not all([X is not None, y is not None, lineages is not None]):
        raise ValueError('Either filename or X, y, and lineages must be provided as input')

    print('Dataset Statistics: ')
    print('Image data shape: ', X.shape)
    print('Number of lineages (should equal batch size): ', len(lineages))

    total_tracks = 0
    total_divisions = 0

    # Calculate cell density
    frame_area = X.shape[2] * X.shape[3]

    avg_cells_in_frame = []
    avg_frame_counts_in_batches = []
    for batch in range(y.shape[0]):
        tracks = lineages[batch]
        total_tracks += len(tracks)
        num_frames_per_track = []

        for cell_lineage in tracks.values():
            num_frames_per_track.append(len(cell_lineage['frames']))
            if cell_lineage.get('daughters', []):
                total_divisions += 1
        avg_frame_counts_in_batches.append(np.average(num_frames_per_track))

        num_cells_in_frame = []
        for frame in range(len(y[batch])):
            y_frame = y[batch, frame]
            cells_in_frame = np.unique(y_frame)
            cells_in_frame = np.delete(cells_in_frame, 0)  # rm background
            num_cells_in_frame.append(len(cells_in_frame))
        avg_cells_in_frame.append(np.average(num_cells_in_frame))

    avg_cells_per_sq_pixel = np.average(avg_cells_in_frame) / frame_area
    avg_num_frames_per_track = np.average(avg_frame_counts_in_batches)

    print('Total number of unique tracks (cells)      - ', total_tracks)
    print('Total number of divisions                  - ', total_divisions)
    print('Average cell density (cells/100 sq pixels) - ', avg_cells_per_sq_pixel * 100)
    print('Average number of frames per track         - ', int(avg_num_frames_per_track))

    return {
        'n_lineages': len(lineages),
        'total_tracks': total_tracks,
        'num_div': total_divisions,
        'avg_cell_density': avg_cells_per_sq_pixel * 100
    }


def get_max_cells(y):
    """Helper function for finding the maximum number of cells in a frame of a movie, across
    all frames of the movie. Can be used for batches/tracks interchangeably with frames/cells.

    Args:
        y (np.array): Annotated image data

    Returns:
        int: The maximum number of cells in any frame
    """
    max_cells = 0
    for frame in range(y.shape[0]):
        cells = np.unique(y[frame])
        n_cells = cells[cells != 0].shape[0]
        if n_cells > max_cells:
            max_cells = n_cells
    return max_cells


def normalize_adj_matrix(adj, epsilon=1e-5):
    """Normalize the adjacency matrix

    Args:
        adj (np.array): Adjacency matrix
        epsilon (float): Used to create the degree matrix

    Returns:
        np.array: Normalized adjacency matrix

    Raises:
        ValueError: If ``adj`` has a rank that is not 3 or 4.
    """
    input_rank = len(adj.shape)
    if input_rank not in {3, 4}:
        raise ValueError('Only 3 & 4 dim adjacency matrices are supported')

    if input_rank == 3:
        # temporarily include a batch dimension for consistent processing
        adj = np.expand_dims(adj, axis=0)

    normalized_adj = np.zeros(adj.shape, dtype='float32')

    for t in range(adj.shape[1]):
        adj_frame = adj[:, t]
        # create degree matrix
        degrees = np.sum(adj_frame, axis=1)
        for batch, degree in enumerate(degrees):
            degree = (degree + epsilon) ** -0.5
            degree_matrix = np.diagflat(degree)

            normalized = np.matmul(degree_matrix, adj_frame[batch])
            normalized = np.matmul(normalized, degree_matrix)
            normalized_adj[batch, t] = normalized

    if input_rank == 3:
        # remove batch axis
        normalized_adj = normalized_adj[0]

    return normalized_adj


def relabel_sequential_lineage(y, lineage):
    """Ensure the lineage information is sequentially labeled.

    Args:
        y (np.array): Annotated z-stack of image labels.
        lineage (dict): Lineage data for y.

    Returns:
        tuple(np.array, dict): The relabeled array and corrected lineage.
    """
    y_relabel, fw, _ = relabel_sequential(y)

    new_lineage = {}

    cell_ids = np.unique(y)
    cell_ids = cell_ids[cell_ids != 0]
    for cell_id in cell_ids:
        new_cell_id = fw[cell_id]

        new_lineage[new_cell_id] = {}

        # Fix label
        # TODO: label == track ID?
        new_lineage[new_cell_id]['label'] = new_cell_id

        # Fix parent
        parent = lineage[cell_id]['parent']
        new_parent = fw[parent] if parent is not None else parent
        new_lineage[new_cell_id]['parent'] = new_parent

        # Fix daughters
        daughters = lineage[cell_id]['daughters']
        new_lineage[new_cell_id]['daughters'] = []
        for d in daughters:
            new_daughter = fw[d]
            if not new_daughter:  # missing labels get mapped to 0
                warnings.warn('Cell {} has daughter {} which is not found '
                              'in the label image `y`.'.format(cell_id, d))
            else:
                new_lineage[new_cell_id]['daughters'].append(new_daughter)

        # Fix frames
        y_true = np.any(y == cell_id, axis=(1, 2))
        y_index = y_true.nonzero()[0]
        new_lineage[new_cell_id]['frames'] = list(y_index)

    return y_relabel, new_lineage


def is_valid_lineage(y, lineage):
    """Check if a cell lineage of a single movie is valid.

    Daughter cells must exist in the frame after the parent's final frame.

    Args:
        y (numpy.array): The 3D label mask.
        lineage (dict): The cell lineages for a single movie.

    Returns:
        bool: Whether or not the lineage is valid.
    """
    all_cells = np.unique(y)
    all_cells = set([c for c in all_cells if c])

    is_valid = True

    # every lineage should have valid fields
    for cell_label, cell_lineage in lineage.items():
        # Get last frame of parent
        if cell_label not in all_cells:
            warnings.warn('Cell {} not found in the label image.'.format(
                cell_label))
            is_valid = False
            continue

        # any cells leftover are missing lineage
        all_cells.remove(cell_label)

        # validate `frames`
        y_true = np.any(y == cell_label, axis=(1, 2))
        y_index = y_true.nonzero()[0]
        frames = list(y_index)
        if frames != cell_lineage['frames']:
            warnings.warn('Cell {} has invalid frames'.format(cell_label))
            is_valid = False
            continue  # no need to test further

        last_parent_frame = cell_lineage['frames'][-1]

        for daughter in cell_lineage['daughters']:
            if daughter not in lineage:
                warnings.warn('Lineage {} has daughter {} not in lineage'.format(
                    cell_label, daughter))
                is_valid = False
                continue  # no need to test further

            # get first frame of daughter
            try:
                first_daughter_frame = lineage[daughter]['frames'][0]
            except IndexError:  # frames is empty?
                warnings.warn('Daughter {} has no frames'.format(daughter))
                is_valid = False
                continue  # no need to test further

            # Check that daughter's start frame is one larger than parent end frame
            if first_daughter_frame - last_parent_frame != 1:
                warnings.warn('Lineage {} has daughter {} in a '
                              'non-subsequent frame.'.format(
                                  cell_label, daughter))
                is_valid = False
                continue  # no need to test further

        # TODO: test parent in lineage
        parent = cell_lineage.get('parent')
        if parent:
            try:
                parent_lineage = lineage[parent]
            except KeyError:
                warnings.warn('Parent {} is not present in the lineage'.format(
                    cell_lineage['parent']))
                is_valid = False
                continue  # no need to test further
            try:
                last_parent_frame = parent_lineage['frames'][-1]
                first_daughter_frame = cell_lineage['frames'][0]
            except IndexError:  # frames is empty?
                warnings.warn('Cell {} has no frames'.format(parent))
                is_valid = False
                continue  # no need to test further
            # Check that daughter's start frame is one larger than parent end frame
            if first_daughter_frame - last_parent_frame != 1:
                warnings.warn(
                    'Cell {} ends in frame {} but daughter {} first '
                    'appears in frame {}.'.format(
                        parent, last_parent_frame, cell_label,
                        first_daughter_frame))
                is_valid = False
                continue  # no need to test further

    if all_cells:  # all cells with lineages should be removed
        warnings.warn('Cells missing their lineage: {}'.format(
            list(all_cells)))
        is_valid = False

    return is_valid  # if unchanged, all cell lineages are valid!


def get_image_features(X, y, appearance_dim=32):
    """Return features for every object in the array.

    Args:
        X (np.array): a 3D numpy array of raw data of shape (x, y, c).
        y (np.array): a 3D numpy array of integer labels of shape (x, y, 1).
        appearance_dim (int): The resized shape of the appearance feature.

    Returns:
        dict: A dictionary of feature names to np.arrays of shape
            (n, c) or (n, x, y, c) where n is the number of objects.
    """
    appearance_dim = int(appearance_dim)

    # each feature will be ordered based on the label.
    # labels are also stored and can be fetched by index.
    num_labels = len(np.unique(y)) - 1
    labels = np.zeros((num_labels,), dtype='int32')
    centroids = np.zeros((num_labels, 2), dtype='float32')
    morphologies = np.zeros((num_labels, 3), dtype='float32')
    appearances = np.zeros((num_labels, appearance_dim,
                            appearance_dim, X.shape[-1]), dtype='float32')

    # iterate over all objects in y
    props = regionprops(y[..., 0], cache=False)
    for i, prop in enumerate(props):
        # Get label
        labels[i] = prop.label

        # Get centroid
        centroid = np.array(prop.centroid)
        centroids[i] = centroid

        # Get morphology
        morphology = np.array([
            prop.area,
            prop.perimeter,
            prop.eccentricity
        ])
        morphologies[i] = morphology

        # Get appearance
        minr, minc, maxr, maxc = prop.bbox
        appearance = np.copy(X[minr:maxr, minc:maxc, :])
        resize_shape = (appearance_dim, appearance_dim)
        appearance = resize(appearance, resize_shape)
        appearances[i] = appearance

    # Get adjacency matrix
    # distance = cdist(centroids, centroids, metric='euclidean') < distance_threshold
    # adj_matrix = distance.astype('float32')

    return {
        'appearances': appearances,
        'centroids': centroids,
        'labels': labels,
        'morphologies': morphologies,
        # 'adj_matrix': adj_matrix,
    }
